subroutine Surface_SHA_mean_qp(lat1, lat2, nlat, lon1, lon2, nlon, f, &
                              &nmax, cs)
!
! ==========================================================================
!
! DESCRIPTION: This subroutine computes spherical harmonic coefficients up to 
!              degree "nmax" by harmonically analysing a global grid of 
!              block-mean values of a signal "f".
!
!              The routine is written in quadruple precision.
!
!              The loop over latitude cells is parallelized using OpenMP.
!
!              The 4*pi fully-normalized associated Legendre functions
!              (e.g., Heiskannen and Moritz, 1967) are evaluated by the
!              numerically stable algorithm by Fukushima (2012). The algorithm
!              can be used up to ultra-high harmonic degrees (say, tens of
!              thousands or even well beyond).
!
!              As the main input, the signal to be analyzed is represented
!              by a matrix of block-mean values as one of the input variables 
!              (see the variable "f" below). Importantly, the whole matrix is 
!              stored in RAM during the entire harmonic analysis. The same 
!              holds for the output spherical harmonic                     
!              coefficients. Ultra-high-degree analysis may therefore
!              require tens of GBs of RAM.
!
!              If possible, the subroutine exploits the symmetry property of 
!              Legendre functions and their integrals with respect to the 
!              equator, thereby improving its performance, especially for 
!              ultra-high-degree harmonic analysis. For further details on the 
!              symmetric grid, see "Surface_SHS_mean_qp.f95".
!
!
! INPUTS: "lat1"      -- Column vector defining the minimum latitudes of the
!                        grid cells in radians (latitudes over a single
!                        meridian).
!
!         "lat2"      -- Column vector defining the maximum latitudes of the
!                        grid cells in radians (latitudes over a single
!                        meridian).
!
!         "nlat"      -- Number of latitudes in the "lat1" and "lat2" arrays.
!
!         "lon1"      -- Column vector defining the minimum longitudes of the
!                        grid cells in radians (longitudes over a single
!                        latitude parallel).
!
!         "lon2"      -- Column vector defining the maximum longitudes of the
!                        grid cells in radians (longitudes over a single
!                        latitude parallel).
!
!         "nlon"      -- Number of longitudes in the "lon1" and "lon2" arrays.
!
!         "f"         -- Input signal, grid of block-mean values.
!
!                        A matrix of the dimensions (nlat, nlon), sampled as 
!                        follows:
!
!
!            [(lat(1,1),lon(1,1))         ...         (lat(1,1),lon(nlon,nlon)]
!            [      .                     .                    .              ]
!            [      .                      .                   .              ]
!            [      .                       .                  .              ]
!            [(lat(nlat,nlat),lon(1,1))   ...   (lat(nlat,nlat),lon(nlon,nlon)]
!
!
!                        where lat(i,i) represents the cell with the minimum 
!                        latitude "lat1(i)" and maximum latitude "lat2(i)" and 
!                        lon(j,j) stands for the the minimum longitude 
!                        "lon1(j)" and maximum longitude "lon2(j)".
!
!         "nmax"    -- Maximum harmonic degree of the output spherical
!                      harmonic coefficients.
!
!
! OUTPUTS: "cs" -- Matrix of dimensions (nmax + 1, nmax + 1) with
!                  output spherical harmonic coefficients. The structure of
!                  of the matrix is as follows:
!
!
!                     [C(0, 0) S(1, 1) S(2, 1)  ...             S(nmax, 1)]
!                     [C(1, 0) C(1, 1) S(2, 2)  ...             S(nmax, 2)]
!                     [C(2, 0) C(2, 1) C(2, 2)  ...             S(nmax, 3)]
!                     [  .       .      .       .                    .    ]
!                     [  .       .      .       .                    .    ]
!                     [  .       .      .        .                   .    ]
!                     [C(nmax, 0)   ...                      C(nmax, nmax)]
!
!
!                  where C(n, m) is the coefficient "C" of degree "n" and order
!                  "m", etc.
!
!
! REFERENCES: Colombo, O. L. (1981) Numerical methods for harmonic analysis on 
!                the sphere. Reports of the Department of Geodetic Science, 
!                Report No. 310, 140 pp
!
!             Sneeuw, N. (1994) Global spherical harmonic analysis by
!                least-squares and numerical quadrature methods in historical
!                perspective. Geophysical Journal International 118:707--716
!
!             Heiskannen, W. A., Moritz, H. (1967) Physical Geodesy. W. H.
!                Freeman and Company, San Francisco, 364 pp
!
!             Jekeli, C., Lee, J. K., Kwon, J. H. (2007) On the computation and 
!               approximation of ultra-high-degree spherical harmonic 
!               series. Journal of Geodesy 81:603--615
!
!             Fukushima, T. (2012) Numerical computation of spherical
!                harmonics of arbitrary degree and order by extending exponent
!                of floating point numbers. Journal of Geodesy 86:271--285.
!
!             Fukushima, T. (2014) Numerical computation of spherical harmonics
!                 of arbitrary degree and order by extending exponent of
!                 floating point numbers: III integral. Computers & Geosciences
!                 63:17--21
!
!
! Contact: blazej.bucha@stuba.sk
!
!
! Please use the following reference when using this subroutine:
!
!          Bucha, B., Hirt, C., Kuhn, M., 2019. Cap integration in
!             spectral gravity forward modelling up to the full gravity
!             tensor. Journal of Geodesy 93:1707--1737
!
!
! Code history: Version 1.0 (Mar 02, 2020)
!
!                           -- The first published version of the package
!
!               Version 1.1 (Apr 04, 2020)
!
!                           -- The package is now provided in double and
!                              quadruple precision.
!
!                           -- The "Surface_SHA_qp.f95" and "Surface_SHA_qp.f95"
!                              routines now employ the FFTW package for FFT. It
!                              was found out that the original routine "FFTCC.f"
!                              (and its quadruple extension) may not yield
!                              sufficiently accurate results in some cases.
!
!                              To run the two routines, the FFTW package has
!                              to already been installed on the user's machine.
!                              The FFTW package then has to be linked during the
!                              compilation.
!
!                           -- A bug has been fixed inside the
!                              "elseif (ixy > 1) then" condition when computing
!                              Legendre functions.
!
! ==========================================================================





    ! Import modules
    ! ===============================================
    use vartypes,  only: qp ! Module defining number formats
    use constants, only: pi_qp, ind_qp, root3_qp, eps_qp ! Module defining
                                                         ! constants                               
                                                         ! (pi, etc.)
    use FFTW3 ! Module for the FFTW package
    ! ===============================================






    ! Declaration of variables
    ! ===============================================
    implicit none

    integer, intent(in)  :: nmax       ! Maximum degree of the output
                                           ! spherical harmonic coefficients
    integer,  intent(in) :: nlat ! Number of grid cells in the latitudinal
                                 ! direction
    integer,  intent(in) :: nlon ! Number of grid cells in the longitudinal
                                 ! direction
    real(qp), intent(in) :: lat1(nlat), lon1(nlon)  ! Minimum spherical
                                                    ! latitudes and longitudes 
                                                    ! of the grid cells in 
                                                    ! radians
    real(qp), intent(in) :: lat2(nlat), lon2(nlon)  ! Maximum spherical
                                                    ! latitudes and longitudes 
                                                    ! of the grid cells in 
                                                    ! radians
    real(qp), intent(in) :: f(nlat, nlon)  ! Input quantity to be harmonically
                                           ! analyzed


    ! Legendre functions (LFs) and some variables necessary for their evaluation
    ! via the extended-range arithmetic (see Fukushima 2012)
    real(qp), parameter :: big = 2.0_qp**ind_qp
    real(qp), parameter :: bigi = 2.0_qp**(-ind_qp)
    real(qp), parameter :: bigs = 2.0_qp**(ind_qp / 2)
    real(qp), parameter :: bigsi = 2.0_qp**(-ind_qp / 2)
    real(qp) :: lat1i, lat2i
    real(qp) :: pnm0_lat1i, pnm0_lat2i   ! LFs
    real(qp) :: pnm1_lat1i, pnm1_lat2i   ! LFs
    real(qp) :: pnm2_lat1i, pnm2_lat2i   ! LFs
    real(qp) :: am(nmax + 1), bm(nmax + 1), dn(nmax + 1)
    real(qp) :: en((nmax + 1) + 1), fn((nmax + 1) + 1)
    real(qp) :: gn(nmax + 1), hn(nmax + 1)
    real(qp) :: ps1((nmax + 1)), ps2((nmax + 1))
    real(qp) :: r(2 * (nmax + 1) + 3), ri(2 * (nmax + 1) + 3)
    real(qp) :: u1, t1, x1, y1, z1, u2, t2, x2, y2, z2, w
    integer  :: ix1, ix2, iy1, iy2, iz1, iz2, ixy1, ixy2
    integer  :: ips1((nmax + 1)), ips2((nmax + 1))
    real(qp) :: in0, imm0, imm1, imm2, inm0, inm1, inm2 ! Integrals of Legendre
                                                        ! functions

    integer :: even    ! If the number of latitudes is even, then "even = 1",
                       ! otherwise "even = 0"
    integer :: symm    ! If the grid is symmetric with respect to the equator,
                       ! then "symm = 1", otherwise "symm = 0". If "symm == 1",
                       ! the subroutine automatically exploits the symmetry
                       ! property of Legendre functions in order to accelerate
                       ! the computation
    integer :: nlatdo  ! If the grid is symmetric with respect to the equator,
                       ! then "nlatdo" represents the number of latitudes from
                       ! one of the poles to the equator (including)
    logical :: symmi   ! Just some auxiliary variable

    integer :: i, j, n, m  ! Loop variables over the latitudes, longitudes,
                           ! harmonic degrees and orders

    real(qp) :: a(nlon / 2 + 1), b(nlon / 2 + 1)   ! Lumped coefficients
    real(qp) :: a2(nlon / 2 + 1), b2(nlon / 2 + 1) ! Lumped coefficients
    real(qp) :: amp1, bmp1    ! Lumped coefficients of order "m"
    real(qp) :: a2mp1, b2mp1  ! Lumped coefficients of order "m"
    real(qp) :: cm, sm
    real(qp) :: lon(nlon) ! "lon1 + lon2"
    real(qp) :: lon0, dlon, dlonchck

    real(qp), parameter :: thold = 10000.0_qp * eps_qp
        ! If abs(a - b) <= thold, then "a" is considered in this subroutine to
        ! be equal to "b" within the threshold "thold". "thold" is used to check
        ! whether input the latitudes and longitudes meet some criteria
        ! (symmetric latitudes with respect to the equator, etc.).
        ! The value of "thold" can be modified by the user if necessary.
        ! The larger the value, the larger numerical inaccuracies will be
        ! tolerated.

    ! Some variables necessary for FFT
    type(c_ptr) :: plan
    real(qp)    :: ftmp_in(nlon)
    complex(qp) :: ftmp_out(nlon / 2 + 1)

    real(qp) :: csatmp, csbtmp  ! Temporary variables for products of LFs and
                                ! lumped coefficients
    real(qp), intent(out) :: cs(nmax + 1, nmax + 1) ! Output matrix of
                                                            ! spherical harmonic
                                                            ! coefficients
    ! ===============================================






    ! The main computational part of the subroutine
    ! ===============================================

    ! Check the latitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Check whether the number of latitudes in "lat1" and "lat2" is even or 
    ! odd
    ! ..........................................................................
    if (mod(nlat, 2) .eq. 0) then
        even = 1  ! The number of cells in the latitudinal direction is an
                  ! even number
    else
        even = 0  ! The number of cells in the latitudinal direction is an
                  ! odd number
    end if
    ! ..........................................................................


    ! Determine whether the cells in the latitudinal direction are symmetric
    ! with respect to the equator
    ! ..........................................................................
    if (nlat == 1) then ! If there is only one cell in the latitudinal
                        ! direction within the grid, the grid is automatically
                        ! considered as not symmetric
        symm = 0
    else
        symm = 1  ! If there is more than one cell in the latidudinal direction
                  ! in the grid, let's start by assuming that the grid is
                  ! symmetric with respect to the equator and check whether
                  ! this is indeed true
    end if


    do i = 1, nlat
        if (abs(lat1(i) + lat2(nlat - i + 1)) > thold) then

            symm = 0  ! The grid is not symmetric.

            exit  ! Exiting the loop with the final decision: the grid is not
                  ! symmetric with respect to the equator ("symm = 0")
        end if
    end do
    ! ..........................................................................


    ! Finally, if the grid is symmetric, we modify the number of latitudes
    ! "nlat" to be equal to the number of latitudes on one hemisphere only
    ! (including the equatorial cells if present). This is because the
    ! time-consuming "do loop" over evaluation points now needs to run for one
    ! hemisphere only, while the results for the other hemishere are obtained
    ! by exploiting the symmetry property of Legendre functions. This reduces
    ! the number of Legendre functions that need to be evaluated by a factor of
    ! ~2, so saves some computational time
    ! ..........................................................................
    if (symm == 1) then
        if (even == 1) then
            nlatdo = nlat / 2
        else
            nlatdo = (nlat + 1) / 2
        end if
    else
        nlatdo = nlat
    end if
    ! ..........................................................................
    ! --------------------------------------------------------------------------



    ! Check the longitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Check whether the first elements of "lon1" and "lon2" are the smallest
    ! one. Otherwise, the execution of the code is terminated.
    if (abs(lon1(1) - minval(lon1)) > thold) then
        error stop '[Surface_SHA_mean_qp.f95 says:] The first element of the &
                  &longitude vector "lon1" has to be the smallest one. &
                  &Terminating the code.'
    end if
    if (abs(lon2(1) - minval(lon2)) > thold) then
        error stop '[Surface_SHA_mean_qp.f95 says:] The first element of the &
                  &longitude vector "lon2" has to be the smallest one. &
                  &Terminating the code.'
    end if


    ! Useful substitution. The variable "lon" represents the mean longitude of 
    ! the grid cells.
    lon = (lon1 + lon2) / 2.0_qp


    ! Get the origin of the longitude "lon" vector (will be necessary later for
    ! the PSLR algorithm)
    lon0 = lon(1) 


    ! Get the lon step of the grid (will be necessary later for the PSLR
    ! algorithm)
    if (nlon > 1) then
        if (abs((lon1(2) - lon1(1)) - (lon2(2) - lon2(1))) > thold) then
            error stop '[Surface_SHA_mean_qp.f95 says:] The difference &
                      &"lon1(2) - lon1(1)" has to be equal to &
                      &"lon2(2) - lon2(1)".'
        end if
        dlon = lon1(2) - lon1(1)
    end if


    ! Check whether the longitude step is constant for the entire "lon1" and
    ! "lon2" vectors. Otherwise, the execution of the code is terminated.
    do j = 2, nlon
        dlonchck = lon1(j) - lon1(j - 1)
        if (abs(dlon - dlonchck) > thold) then
            error stop '[Surface_SHA_mean_qp.f95 says:] The step of "lon1" &
                  &has to be constant. Terminating the code.'
        end if

        dlonchck = lon2(j) - lon2(j - 1)
        if (abs(dlon - dlonchck) > thold) then
            error stop '[Surface_SHA_mean_qp.f95 says:] The step of "lon2" &
                  &has to be constant. Terminating the code.'
        end if
    end do
    ! --------------------------------------------------------------------------



    ! Initializations for recurrence relations to compute Legendre functions
    ! --------------------------------------------------------------------------
    do m = 1, (2 * (nmax + 1) + 3)
        w = sqrt(real(m, qp))
        r(m) = w
        ri(m) = 1.0_qp / w
    end do


    ! Computation of the "dn" coefficients for Legendre recurrence relations
    do n = 1, (nmax + 1)
        dn(n) = r(2 * n + 3) * ri(2 * n + 2)
    end do


    ! Coefficients "en" and "fn" for recurrence relations to compute 
    ! *un-normalized* Legendre polynomials. (Note that to compute integrals of 
    ! fully-normalized Legendre polynomials, we need un-normalized Legendre 
    ! polynomials; see, e.g., Eq. B.24 of Jekeli et al. 2007)
    do n = 2, (nmax + 1)
        en(n + 1) = real(2 * n - 1, qp) / real(n, qp)
        fn(n + 1) = real(n - 1, qp) / real(n, qp)
    end do
    ! --------------------------------------------------------------------------



    ! Initializations for recurrence relations to compute integrals of 
    ! fully-normalized Legendre functions
    ! --------------------------------------------------------------------------
    ! Computation of "gn" coefficients
    do n = 3, nmax
        gn(n + 1) = 1.0_qp / real(2 * n + 2, qp) * r(n) * r(2 * n + 1) * &
                    r(2 * n - 1) * ri(n - 1)
    end do

    ! Computation of "hn" coefficients
    do n = 0, nmax
        hn(n + 1) = real(n - 2, qp) / real(n + 1, qp)
    end do
    ! -------------------------------------------------------------------------- 



    ! Initialize all elements of the output coefficients to zero
    ! --------------------------------------------------------------------------
    cs = 0.0_qp
    ! --------------------------------------------------------------------------






    ! Create a plan for FFT
    ! --------------------------------------------------------------------------
    plan = fftwq_plan_dft_r2c_1d(nlon, ftmp_in, ftmp_out, FFTW_ESTIMATE)
    ! --------------------------------------------------------------------------






    ! Loop over latitudes
    ! --------------------------------------------------------------------------
    !$OMP PARALLEL DO REDUCTION(+:cs) DEFAULT(NONE) &
    !$OMP& SHARED(f, lat1, lat2, nlat, dlon, w, nlatdo, even, nmax) &
    !$OMP& SHARED(dn, en, fn, hn, gn, r, ri, plan, symm) &
    !$OMP& PRIVATE(lat1i, lat2i) &
    !$OMP& PRIVATE(pnm0_lat1i, pnm1_lat1i, pnm2_lat1i, t1, u1) &
    !$OMP& PRIVATE(pnm0_lat2i, pnm1_lat2i, pnm2_lat2i, t2, u2) &
    !$OMP& PRIVATE(x1, ix1, y1, iy1, z1, iz1, ps1, ips1, ixy1) &
    !$OMP& PRIVATE(x2, ix2, y2, iy2, z2, iz2, ps2, ips2, ixy2) &
    !$OMP& PRIVATE(a, b, a2, b2, am, bm, amp1, a2mp1, bmp1, b2mp1) &
    !$OMP& PRIVATE(in0, inm0, inm1, inm2, imm0, imm1, imm2) &
    !$OMP& PRIVATE(csatmp, csbtmp, cm, sm, ftmp_in, ftmp_out, symmi)
    do i = 1, nlatdo

        ! Check whether the symmetry property of LFs needs to be exploited
        ! ------------------------------------------------------------------
        if (nlat == 1) then ! There is only a single latitude parallel
                            ! in the grid, so no symmetry property
            symmi = .false.
        elseif (symm == 0) then ! The grid was already classified as
                                ! non-symmetric
            symmi = .false.
        elseif (symm == 1 .and. even == 0 .and. i == nlatdo) then
            symmi = .false.  ! For a symmetric grid containing the equator,
                             ! this ensures that the points on the equator
                             ! will not be processed twice
        else
            symmi = .true. ! Exploit the symmetry property
        end if
        ! ------------------------------------------------------------------


        ! Lumped coefficients for the southern hemisphere (including the
        ! equatorial cells if any)
        ! ------------------------------------------------------------------
        ftmp_in = f(i, :)
        call fftwq_execute_dft_r2c(plan, ftmp_in, ftmp_out)

        a  =  real(ftmp_out, qp)
        b  = -aimag(ftmp_out)
        ! ------------------------------------------------------------------


        ! Lumped coefficients for the northern hemisphere
        ! ------------------------------------------------------------------
        if (symmi .eqv. .true.) then
            ftmp_in = f(nlat - i + 1, :)
            call fftwq_execute_dft_r2c(plan, ftmp_in, ftmp_out)

            a2 =  real(ftmp_out, qp)
            b2 = -aimag(ftmp_out)
        end if
        ! ------------------------------------------------------------------


        ! Some precomputations for Legendre functions
        ! ------------------------------------------------------------------
        ! lat1(i)
        ! ..................................................................
        lat1i = lat1(i)
        t1  = sin(lat1i)
        u1  = cos(lat1i)
        x1  = root3_qp * u1
        ix1 = 0

         ps1(1) =  x1
        ips1(1) = ix1

        do n = 2, (nmax + 1)
            x1 = (dn(n - 1) * u1) * x1
            y1 = abs(x1)

            if (y1 >= bigs) then
                 x1 =  x1 * bigi
                ix1 = ix1 + 1
            elseif (y1 < bigsi) then
                 x1 =  x1 * big
                ix1 = ix1 - 1
            end if

             ps1(n) =  x1
            ips1(n) = ix1
        end do
        ! ..................................................................


        ! lat2(i)
        ! ..................................................................
        lat2i = lat2(i)
        t2  = sin(lat2i)
        u2  = cos(lat2i)
        x2  = root3_qp * u2
        ix2 = 0

         ps2(1) =  x2
        ips2(1) = ix2

        do n = 2, (nmax + 1)
            x2 = (dn(n - 1) * u2) * x2
            y2 = abs(x2)

            if (y2 >= bigs) then
                 x2 =  x2 * bigi
                ix2 = ix2 + 1
            elseif (y2 < bigsi) then
                 x2 =  x2 * big
                ix2 = ix2 - 1
            end if

             ps2(n) =  x2
            ips2(n) = ix2
        end do
        ! ..................................................................
        ! ------------------------------------------------------------------






        ! Loop over harmonic orders
        ! ------------------------------------------------------------------
        do m = 0, nmax


            ! Due to the use of the mean values, some additional terms need to 
            ! be taken into account when compared with the harmonic analysis 
            ! based on point values (see, e.g., Colombo, 1981)
            ! ------------------------------------------------------------------
            if (m == 0) then
                amp1  = a(m + 1) * dlon

                if (symmi .eqv. .true.) then
                    a2mp1 = a2(m + 1) * dlon
                end if
            else
                ! Useful substitution
                cm = (cos(real(m, qp) * dlon) - 1.0_qp) / real(m, qp)
                sm = sin(real(m, qp) * dlon) / real(m, qp)

                amp1  =  a(m + 1) * sm + b(m + 1) * cm
                bmp1  = -a(m + 1) * cm + b(m + 1) * sm 

                if (symmi .eqv. .true.) then
                    a2mp1 =  a2(m + 1) * sm + b2(m + 1) * cm
                    b2mp1 = -a2(m + 1) * cm + b2(m + 1) * sm
                end if
            end if
            ! ------------------------------------------------------------------


            ! Computation of "am" and "bm" coefficients for Legendre recurrence
            ! relations
            ! ------------------------------------------------------------------
            am(m + 1) = r(2 * m + 3)
            do n = (m + 2), (nmax + 1) 
                w     = r(2 * n + 1) * ri(n - m) * ri(n + m)
                am(n) = r(2 * n - 1) * w
                bm(n) = r(n - m - 1) * r(n + m - 1) * ri(2 * n - 3) * w
            end do
            ! ------------------------------------------------------------------


            ! Computation of spherical harmonic coefficients
            ! ------------------------------------------------------------------
            if (m == 0) then






                ! Zonal Legendre functions and their integrals
                ! --------------------------------------------------------------
                ! P00
                ! .............................................................
                pnm0_lat1i = 1.0_qp
                pnm0_lat2i = 1.0_qp
                ! .............................................................


                ! P10
                ! .............................................................
                pnm1_lat1i = t1
                pnm1_lat2i = t2
                ! .............................................................


                ! I00
                ! .............................................................
                in0 = t2 - t1
                ! .............................................................


                ! Spherical harmonic coefficients
                ! .............................................................
                csatmp = in0 * amp1
                if (symmi .eqv. .true.) then
                    csatmp = csatmp + in0 * a2mp1
                end if
                cs(1, m + 1) = cs(1, m + 1) + csatmp ! C00
                ! .............................................................


                if ((nmax + 1) >= 2) then

                    do n = 1, nmax
                        ! P20, P30, ..., Pnmax+1,0
                        ! Since we use locally the "n" variable to compute 
                        ! Legendre polynomials of degree "n + 1", the "n + 2"th 
                        ! elements have to be taken from the vectors "en" and 
                        ! "fn" below. Once again, note that when "m == 0", then 
                        ! "pnm0", "pnm1" and "pnm2" for lat1i and lat2i 
                        ! represent un-normalized Legendre polynomials. These 
                        ! are needed to get the integrals of fully-normalized 
                        ! Legendre polynomials
                        ! .....................................................
                        pnm2_lat1i = en(n + 2) * t1 * pnm1_lat1i - &
                                     fn(n + 2) * pnm0_lat1i
                        pnm2_lat2i = en(n + 2) * t2 * pnm1_lat2i - &
                                     fn(n + 2) * pnm0_lat2i
                        ! .....................................................


                        ! I10, I20, ... Inmax,0
                        ! Computed from Pn+1,0 and Pn-1,0
                        ! .....................................................
                        in0 = ri(2 * n + 1) * (pnm2_lat2i - pnm0_lat2i - &
                                               pnm2_lat1i + pnm0_lat1i)
                        ! .....................................................


                        ! Spherical harmonic coefficients
                        ! .....................................................
                        csatmp = in0 * amp1

                        if (symmi .eqv. .true.) then
                            if (mod(n, 2) == 0) then
                                csatmp = csatmp + in0 * a2mp1
                            else
                                csatmp = csatmp - in0 * a2mp1
                            end if
                        end if
                        cs(n + 1, m + 1) = cs(n + 1, m + 1) + csatmp ! C10, 
                                                                     ! C20,..., 
                                                                     ! Cnmax,0
                        ! .....................................................


                        pnm0_lat1i = pnm1_lat1i
                        pnm1_lat1i = pnm2_lat1i


                        pnm0_lat2i = pnm1_lat2i
                        pnm1_lat2i = pnm2_lat2i

                    end do

                end if
                ! --------------------------------------------------------------
                





            else ! Non-zonal harmonics






                ! Sectorial Legendre functions and their integrals
                ! --------------------------------------------------------------

                ! Pmm for lat1i
                ! .............................................................
                x1  =  ps1(m)
                ix1 = ips1(m)
    
                if (ix1 == 0) then
                    pnm0_lat1i = x1
                elseif (ix1 < -1) then
                    pnm0_lat1i = 0.0_qp
                elseif (ix1 < 0) then
                    pnm0_lat1i = x1 * bigi
                else
                    pnm0_lat1i = x1 * big
                end if
                ! .............................................................


                ! Pmm for lat2i
                ! .............................................................
                x2  =  ps2(m)
                ix2 = ips2(m)
    
                if (ix2 == 0) then
                    pnm0_lat2i = x2
                elseif (ix2 < -1) then
                    pnm0_lat2i = 0.0_qp
                elseif (ix2 < 0) then
                    pnm0_lat2i = x2 * bigi
                else
                    pnm0_lat2i = x2 * big
                end if
                ! .............................................................


                ! Imm
                ! .............................................................
                if (m == 1) then ! I11

                    imm0 = sqrt(3.0_qp) / 2.0_qp * &
                          ((t2 * u2 - (pi_qp / 2.0_qp - lat2i)) - &
                           (t1 * u1 - (pi_qp / 2.0_qp - lat1i)))
                    inm0 = imm0

                elseif (m == 2) then ! I22

                    imm1 = sqrt(15.0_qp) / 6.0_qp * (t2 * (3.0_qp - t2 * t2) -&
                                                     t1 * (3.0_qp - t1 * t1))
                    inm0 = imm1

                else ! I33, I44, ..., Inmax,nmax
                    
                    imm2 = gn(m + 1) * imm0 + 1.0_qp / real(m + 1, qp) * &
                           (t2 * pnm0_lat2i - t1 * pnm0_lat1i)
                    inm0 = imm2
    
                    imm0 = imm1
                    imm1 = imm2

                end if
                ! .............................................................


                ! Spherical harmonic coefficients
                ! .............................................................
                csatmp = inm0 * amp1
                csbtmp = inm0 * bmp1

                if (symmi .eqv. .true.) then
                    csatmp = csatmp + inm0 * a2mp1
                    csbtmp = csbtmp + inm0 * b2mp1
                end if
                cs(m + 1, m + 1) = cs(m + 1, m + 1) + csatmp ! Cmm
                cs(m, m + 1)     = cs(m, m + 1)     + csbtmp ! Smm
                ! .............................................................
                ! --------------------------------------------------------------






                ! Tesseral harmonics
                ! --------------------------------------------------------------
                if (m < nmax) then

                    ! Pm+1,m for lat1i
                    ! .........................................................
                     y1 =  x1
                    iy1 = ix1
                     x1 = (am(m + 1) * t1) * y1
                    ix1 = iy1
    
                    w = abs(x1)
                    if (w >= bigs) then
                         x1 =  x1 * bigi
                        ix1 = ix1 + 1
                    elseif (w < bigsi) then
                         x1 =  x1 * big
                        ix1 = ix1 - 1
                    end if
    
                    if (ix1 == 0) then
                        pnm1_lat1i = x1
                    elseif (ix1 < - 1) then
                        pnm1_lat1i = 0.0_qp
                    elseif (ix1 < 0) then
                        pnm1_lat1i = x1 * bigi
                    else
                        pnm1_lat1i = x1 * big
                    end if
                    ! .........................................................



                    ! Pm+1,m for lat2i
                    ! .........................................................
                     y2 =  x2
                    iy2 = ix2
                     x2 = (am(m + 1) * t2) * y2
                    ix2 = iy2
    
                    w = abs(x2)
                    if (w >= bigs) then
                         x2 =  x2 * bigi
                        ix2 = ix2 + 1
                    elseif (w < bigsi) then
                         x2 =  x2 * big
                        ix2 = ix2 - 1
                    end if
    
                    if (ix2 == 0) then
                        pnm1_lat2i = x2
                    elseif (ix2 < - 1) then
                        pnm1_lat2i = 0.0_qp
                    elseif (ix2 < 0) then
                        pnm1_lat2i = x2 * bigi
                    else
                        pnm1_lat2i = x2 * big
                    end if
                    ! .........................................................


                    ! Im+1,m
                    ! .........................................................
                    ! This is not a typo, "pnm0" are indeed required here
                    inm1 = -am(m + 1) / real(m + 2, qp) * &
                            (u2 * u2 * pnm0_lat2i - &
                             u1 * u1 * pnm0_lat1i)    
                    ! .........................................................


                    ! Spherical harmonic coefficients
                    ! .........................................................
                    csatmp = inm1 * amp1
                    csbtmp = inm1 * bmp1

                    if (symmi .eqv. .true.) then
                        csatmp = csatmp - inm1 * a2mp1
                        csbtmp = csbtmp - inm1 * b2mp1
                    end if
                    cs(m + 2, m + 1) = cs(m + 2, m + 1) + csatmp ! Cm+1,m
                    cs(m, m + 2)     = cs(m, m + 2)     + csbtmp ! Sm+1,m
                    ! .........................................................


                    ! Pm+2,m, Pm+3,m, ..., Pnmax,m and their integrals
                    ! .........................................................
                    do n = (m + 2), nmax

                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m for lat1i
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ixy1 = ix1 - iy1
    
                        if (ixy1 == 0) then
                             z1 = (am(n) * t1) * x1 - bm(n) * y1
                            iz1 = ix1
                        elseif (ixy1 == 1) then
                             z1 = (am(n) * t1) * x1 - bm(n) * (y1 * bigi)
                            iz1 = ix1
                        elseif (ixy1 == -1) then
                             z1 = (am(n) * t1) * (x1 * bigi) - bm(n) * y1
                            iz1 = iy1
                        elseif (ixy1 > 1) then
                             z1 = (am(n) * t1) * x1
                            iz1 = ix1
                        else
                             z1 = -bm(n) * y1
                            iz1 = iy1
                        end if
    
                        w = abs(z1)
                        if (w >= bigs) then
                             z1 =  z1 * bigi
                            iz1 = iz1 + 1
                        elseif (w < bigsi) then
                             z1 =  z1 * big
                            iz1 = iz1 - 1
                        end if
    
                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz1 == 0) then
                            pnm2_lat1i = z1
                        elseif (iz1 < -1) then
                            pnm2_lat1i = 0.0_qp
                        elseif (iz1 < 0 ) then
                            pnm2_lat1i = z1 * bigi
                        else
                            pnm2_lat1i = z1 * big
                        end if
    
                         y1 =  x1
                        iy1 = ix1
                         x1 =  z1
                        ix1 = iz1
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m for lat2i
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ixy2 = ix2 - iy2
    
                        if (ixy2 == 0) then
                             z2 = (am(n) * t2) * x2 - bm(n) * y2
                            iz2 = ix2
                        elseif (ixy2 == 1) then
                             z2 = (am(n) * t2) * x2 - bm(n) * (y2 * bigi)
                            iz2 = ix2
                        elseif (ixy2 == -1) then
                             z2 = (am(n) * t2) * (x2 * bigi) - bm(n) * y2
                            iz2 = iy2
                        elseif (ixy2 > 1) then
                             z2 = (am(n) * t2) * x2
                            iz2 = ix2
                        else
                             z2 = -bm(n) * y2
                            iz2 = iy2
                        end if
    
                        w = abs(z2)
                        if (w >= bigs) then
                             z2 =  z2 * bigi
                            iz2 = iz2 + 1
                        elseif (w < bigsi) then
                             z2 =  z2 * big
                            iz2 = iz2 - 1
                        end if
    
                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz2 == 0) then
                            pnm2_lat2i = z2
                        elseif (iz2 < -1) then
                            pnm2_lat2i = 0.0_qp
                        elseif (iz2 < 0 ) then
                            pnm2_lat2i = z2 * bigi
                        else
                            pnm2_lat2i = z2 * big
                        end if
    
                         y2 =  x2
                        iy2 = ix2
                         x2 =  z2
                        ix2 = iz2
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Im+2,m, Im+3,m, ..., Inmax,m
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ! This is not a typo, "pnm1" are indeed required here
                        inm2 = hn(n + 1) * bm(n) * inm0 - &
                               am(n) / real(n + 1, qp) * &
                               (u2 * u2 * pnm1_lat2i - u1 * u1 * pnm1_lat1i)

                        inm0 = inm1
                        inm1 = inm2
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Spherical harmonic coefficients
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        csatmp = inm2 * amp1
                        csbtmp = inm2 * bmp1

                        if (symmi .eqv. .true.) then

                            if (mod(n + m, 2) == 0) then
                                csatmp = csatmp + inm2 * a2mp1
                                csbtmp = csbtmp + inm2 * b2mp1
                            else
                                csatmp = csatmp - inm2 * a2mp1
                                csbtmp = csbtmp - inm2 * b2mp1
                            end if

                        end if


                        cs(n + 1, m + 1) = cs(n + 1, m + 1) + csatmp ! Cm+2,m,
                                                                     ! Cm+3,m,
                                                                     ! ...,
                                                                     ! Cnmax,m
                        cs(m, n + 1)     = cs(m, n + 1)     + csbtmp ! Sm+2,m,
                                                                     ! Sm+3,m,
                                                                     ! ...,
                                                                     ! Snmax,m
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        pnm1_lat1i = pnm2_lat1i
                        pnm1_lat2i = pnm2_lat2i

                    end do ! End of the loop over harmonic degrees
                    ! ----------------------------------------------------------

                end if ! End of computation of tesseral harmonics
                ! --------------------------------------------------------------

            end if ! End of computation of spherical harmonic coefficients
            ! ------------------------------------------------------------------

        end do ! End of the loop over harmonic orders
        ! ------------------------------------------------------------------

    end do ! End of the loop over grid latitudes
    !$OMP END PARALLEL DO

    call fftwq_destroy_plan(plan)
    ! ------------------------------------------------------------------

    cs = cs / (4.0_qp * pi_qp)
    ! ===============================================

end subroutine Surface_SHA_mean_qp

